#!python3

import sys
import re
import argparse
import subprocess as sp
from datetime import datetime
from collections import defaultdict

class logger:
    def __init__(self, release, annotation, logname='changelog.md', internal=False):
        self.release = release
        self.annotation = annotation
        self.logname = logname
        self.internal = internal

    def make(self):
        self.__redirect_output(self.logname)
        log = self.__prepare_log()
        changelog = self.__parse_git_log(log, self.release, self.annotation)
        self.__print_changelog(changelog, self.internal)

    def __redirect_output(self, filename):
        try:
            sys.stdout = open(filename, 'w')
        except Exception as e:
            print('unable to open: ', e.args)

    def __prepare_log(self, ):
        log_proc = sp.run('git log --date=format:\'%Y.%m.%d\' --pretty=format:\'%s | %D | %ad\'',
            shell=True, stdout=sp.PIPE, universal_newlines=True);
        code=log_proc.returncode
        if code:
            print(f'git returned error code: {code}')
            exit(code)
        #list of lines
        log=log_proc.stdout.split('\n')
        #list of lists of fields
        #[[comment, relog, date]]
        log=list(map(lambda line: line.split('|'), log))
        return log

    def __parse_comment(self, changelog, tag, comment):
        tokens=['[feature]', '[fix]', '[changelog]', '[internal]']
        comment = comment.split(' ')
        comment = list(filter(lambda word: len(word), comment))
        token='[else]' #everything else falls in [else] section.
        description = ''
        for word in comment:
            if word in tokens:
                if len(description):
                    changelog[tag][token].append(description)
                token = word
                description = ''
            else:
                description = description +' '+word
        if len(description):
            changelog[tag][token].append(description)


    def __parse_git_log(self, log_out, tagname='unreleased', summary='no summary'):
        #until first valid tag in timeline
        #everything is unreleased
        changelog=defaultdict(lambda: defaultdict(list))
        tag=tagname
        changelog[tag]['summary'] = summary.split('\n')
        changelog[tag]['date'] = datetime.now().strftime('%Y.%m.%d')
        for log in log_out:
            comment, reflog, date = log
            tagmatch = re.search(r'(?<=tag: )([\w.]+)', reflog)
            if tagmatch:
                #there is new release tag.
                tag=tagmatch.group(0)
                #add tag annotation for release
                tag_proc = sp.run(f'git for-each-ref refs/tags/{tag} --format=\'%(contents)\'',
                    shell=True, stdout=sp.PIPE, universal_newlines=True);
                tagsum=tag_proc.stdout.split('\n')
                changelog[tag]['summary'] = tagsum
                changelog[tag]['date'] = date
            self.__parse_comment(changelog, tag, comment)
        return changelog

    def __print_section(self, d, key, section_text):
        if len(d[key]):
            print(section_text)
            for feat in d[key]:
                print(f' - [x] {feat}')

    def __print_summary(self, d):
        summary = list(filter(lambda line: len(line), d['summary']))
        if len(summary):
            print("```")
            for sumline in d['summary']:
                if len(sumline):
                    print(sumline)
            print("```")

    def __print_changelog(self, changelog, is_internal=False):
        for k, v in changelog.items():
            print('\n---\n')
            print(f'# {k} ({v["date"]} )\n')
            self.__print_summary(v)
            self.__print_section(v, '[feature]', '### new features:')
            self.__print_section(v, '[fix]', '### bug fixes:')
            self.__print_section(v, '[changelog]', '### minor changes:')
            if is_internal:
                self.__print_section(v, '[internal]', '### for developers only:')

                


from .logger import logger
import argparse

description = '''
This script will make changelog,
for git tracked projects
'''
epilog='''
\x1b[36m
commit message format:
[feature]|[fix]|[changelog] actual_commit_message
[feature] stands for new major features introduced by the commit
[fix] for major buxfixes made by commit
[changelog] else additional helpfull information for changelog
[internal] messages used only for internal changelogs. These messages can be
    ommited for your customers with flag '--internal yes'.
\x1b[0m'''
parser = argparse.ArgumentParser(description=description,
    epilog=epilog, formatter_class=argparse.RawDescriptionHelpFormatter)
parser.add_argument('--logname', default='changelog.md', type=str,
    help='Filename for the changelog')
parser.add_argument('--internal', default=False, type=bool,
    help='If internal is False, [internal] commit messages are ommited.')
parser.add_argument('release',  type=str,
    help='Tagname for the unreleased commits.')
parser.add_argument('annotation', type=str,
    help='Annotation for the release. Must contain summary information.')
args=parser.parse_args()

l = logger(args.release, args.annotation, args.logname, args.internal)
l.make()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['changelog_md']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'changelog-md',
    'version': '0.1.3',
    'description': 'A package to create changelogs for your git tracked projects.',
    'long_description': '# install with pip\n```bash\npython3 -m pip install changelog-md\n``` \n\n# install with Poetry\n```\npoetry add changelog-md\n```\n\n# execution information:\n\nThis script aim to provide changelogs on git-tracked projects.\nRun `changelog.py` script in your directory \nand it\'ll generate the changelog file.\nIn order to obtain actual changelog, run:\n\n\nYou can use git changelog from command line directly:\n```bash\n# ... code on project\ngit commit                                                    #add your commit message on work you\'ve done\npython3 -m changelog-md vX.Y.Z "Annotation for this version" #generate changelog\ngit add changelog.md                                          #stage the changelog to current commit\ngit commit --amend                                            #add the changelog to current commit, leave commit message as is.\ngit tag -a vX.Y.Z -m "Annotation for this version"            #add the tag you\'ve set earlier in changelog.\ngit push --follow-tags                                        #publish your changes and the new tag.\n```\n\nOr you can elaborate your workflow python scripts \nto use changelog-md as a python module:\n```python\n# CICD_script.py\n# ... do some CICD work\n\nfrom changelog-md import logger\nl = logger(\'vX.Y.Z\', \'Annotation for this version\')           #create logger instance\nl.make()                                                      #print the log\n\n# ... do other CICD work\n```\n\n\nMore actual information on colophon of scripts:\n```shell\n#actual information on colophon:\npython3 -m changelog-md --help\n```\n\n# correct commit messages:\n\nTo make this scripts work properly, your commit messages\nshould look like any of this:\n```\n[feature] feature description here\n[fix] fix description\n[changelog] minor changes description\n[internal] this section will be generated only if --internal argument is given.\nThis commit message string will be ommited from changelog.\n```\n\n#  Releases:\n\nAny tagged commit will be interpreted as release.\nTo create tag, write:\n```\ngit tag -a v<version>\n... tag anotation ...\ngit push --follow-tags\n```\nAny commit above the last tag will be marked as tag provided \nin command line arguments and annotated accordingly.\n',
    'author': 'Ruslan Sergeev',
    'author_email': 'mybox.sergeev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/RuslanSergeev/git_changelog',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
